# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test for the virtual side creator.
"""
import pytest
import vermouth
from vermouth.rcsu.go_vs_includes import VirtualSiteCreator
from vermouth.tests.helper_functions import test_molecule

def test_no_moltype_error(test_molecule):
    """
    Test that various high level errors are
    properly raised.
    """
    # set up processor
    processor = VirtualSiteCreator()
    # no moltype set
    system = vermouth.System()
    system.add_molecule(test_molecule)
    with pytest.raises(ValueError):
        processor.run_system(system)

def test_no_system_error(test_molecule):
    """
    Test that various high level errors are
    properly raised.
    """
    # set up processor
    processor = VirtualSiteCreator()
    test_molecule.meta['moltype'] = "random"
    # no system
    with pytest.raises(ValueError):
        processor.run_molecule(test_molecule)

def test_return_no_nodes():
    mol = vermouth.Molecule()
    processor = VirtualSiteCreator()
    mol.meta['moltype'] = "random"
    system = vermouth.System()
    system.add_molecule(mol)
    assert processor.run_system(system) is None



"""
Test for the water bias processor.
"""
import pytest
import numpy as np
import networkx as nx
import vermouth
from vermouth.rcsu.go_vs_includes import VirtualSiteCreator
from vermouth.tests.test_water_bias import create_sys_all_attrs
from vermouth.tests.helper_functions import test_molecule
from vermouth.rcsu.go_structure_bias import ComputeStructuralGoBias

def test_compute_go_interaction(test_molecule):
    contacts = [("atom_a", "atom_b", 2.3),
                ("atom_a", "atom_c", 4.1),
                ("atom_q", "atom_p", 0.0)]
    denom = 2**(1/6.)
    expected = {(k1, k2): v/denom for k1, k2, v in contacts}


    system = vermouth.System()
    system.add_molecule(test_molecule)
    go_processor = ComputeStructuralGoBias(contact_map=None,
                                           cutoff_short=None,
                                           cutoff_long=None,
                                           go_eps=2.1,
                                           res_dist=None,
                                           moltype="molecule_0")
    go_processor.system = system

    go_processor.compute_go_interaction(contacts)
    for nb_params in system.gmx_topology_params['nonbond_params']:
        assert nb_params.atoms in expected
        assert pytest.approx(expected[nb_params.atoms]) == nb_params.sigma
        assert nb_params.epsilon == 2.1

@pytest.mark.parametrize('cmap, cshort, clong, rdist, expected',(
        # single assymetric contact bad
        ([(1, 'A', 4, 'A')],
         0.3,
         2.0,
         0,
         []),
        # single symmetric contact good
        ([(1, 'A', 4, 'A'), (4, 'A', 1, 'A')],
         0.3,
         2.0,
         0,
         [("mol_0_4", "mol_0_1", 1.5)]),
        # single symmetric contact bad -> cshort
        ([(3, 'A', 1, 'A'),(1, 'A', 3, 'A')],
         1.5,
         2.0,
         0,
         []),
        # single symmetric contact bad -> clong
        ([(4, 'A', 1, 'A'), (1, 'A', 4, 'A')],
         0.5,
         0.8,
         0,
         []),
        # single symmetric contact bad -> rdist
        ([(4, 'A', 1, 'A'), (1, 'A', 4, 'A')],
         0.3,
         2.0,
         5,
         []),
        # single contact bad -> cshort & assym
        ([(1, 'A', 3, 'A')],
         1.5,
         2.0,
         0,
         []),
        # single contact bad -> clong & assym
        ([(1, 'A', 4, 'A')],
         0.5,
         0.8,
         0,
         []),
        # single contact bad -> rdist & assym
        ([(1, 'A', 4, 'A')],
         0.3,
         2.0,
         5,
         []),
        # two symmetric contacts good
        ([(3, 'A', 1, 'A'), (4, 'A', 1, 'A'),(1, 'A', 3, 'A'), (1, 'A', 4, 'A')],
         0.3,
         2.0,
         1,
         [("mol_0_1", "mol_0_3", 1.0),
          ("mol_0_1", "mol_0_4", 1.5),
         ]),
        # one symmetric contacts good, one assymetric contact bad
        ([(3, 'A', 1, 'A'), (4, 'A', 1, 'A'),(1, 'A', 3, 'A')],
         0.3,
         2.0,
         1,
         [("mol_0_1", "mol_0_3", 1.0)]),
        # one good one bad rdist
        ([(4, 'A', 1, 'A'), (2, 'A', 1, 'A'), (1, 'A', 4, 'A'), (1, 'A', 2, 'A')],
         0.3,
         2.0,
         2,
         [("mol_0_1", "mol_0_4", 1.5)]),
        ))
def test_contact_selector(test_molecule, 
                          cmap, 
                          cshort, 
                          clong, 
                          rdist, 
                          expected):

    # define some possible positions
    pos = np.array([[0.0, 0.0, 0.0],
                    [0.0, 0.4, 0.0],
                    [0.0, 0.8, 0.0],
                    [0.5, 0.0, 0.0],
                    [0.5, 0.4, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.5, 0.0, 0.0],
                    [1.5, 0.4, 0.0],
                    [1.5, 0.8, 0.0]])
    for node in test_molecule.nodes:
        test_molecule.nodes[node]['position'] = pos[node]

    # bead sizes
    sizes = {0: 0.47, 3: 0.41, 5: 0.38, 6: 0.47}
    # the molecule atomtypes
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a", 
              3: "SP1", 4: "C1",
              5: "TP1", 
              6: "P1", 7: "SN3a", 8: "SP4"}
    # the molecule resnames
    resnames = {0: "A", 1: "A", 2: "A", 
                3: "B", 4: "B",
                5: "C", 
                6: "D", 7: "D", 8: "D"}

    secstruc = {1: "H", 2: "H", 3: "H", 4: "H"}
    system = create_sys_all_attrs(test_molecule, 
                                  moltype="mol_0", 
                                  secstruc=secstruc,
                                  defaults={"chain": "A"}, 
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    # generate the virtual sites
    VirtualSiteCreator().run_system(system)
    # initialize the Go processor
    go_processor = ComputeStructuralGoBias(contact_map=cmap,
                                           cutoff_short=cshort,
                                           cutoff_long=clong,
                                           go_eps=2.1,
                                           res_dist=rdist,
                                           moltype="mol_0")
    go_processor.res_graph = vermouth.graph_utils.make_residue_graph(test_molecule)
    # run the contact map selector
    contact_matrix = go_processor.contact_selector(test_molecule)
    assert contact_matrix == expected



# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test for the tune idp bonds processor.
"""
import pytest
from vermouth.processors.annotate_idrs import AnnotateIDRs
from vermouth.tests.helper_functions import create_sys_all_attrs, test_molecule

@pytest.mark.parametrize('idr_regions, expected', [
    (
    [(1,2)],
    [True, True, True, True, True, False, False, False, False]
    ),
    (
    [],
    [False, False, False, False, False, False, False, False, False]
    )
])
def test_make_disorder_string(test_molecule,
                              idr_regions,
                              expected):
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}
    # the molecule resnames
    resnames = {0: "A", 1: "A", 2: "A",
                3: "B", 4: "B",
                5: "C",
                6: "D", 7: "D", 8: "D"}
    secstruc = {1: "H", 2: "H", 3: "H", 4: "H"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    AnnotateIDRs(id_regions=idr_regions).run_system(system)
    result = []
    for key, node in system.molecules[0].nodes.items():
        if system.molecules[0].nodes[key].get("cgidr"):
            result.append(True)
        else:
            result.append(False)
    assert result == expected



# Copyright 2023 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Obtain the structural bias for the Go model.
"""
import numpy as np
import networkx as nx
from ..graph_utils import make_residue_graph
from ..molecule import Interaction
from ..processors.processor import Processor
from ..selectors import filter_minimal, select_backbone
from ..gmx.topology import NonbondParam
from .go_utils import get_go_type_from_attributes

class ComputeStructuralGoBias(Processor):
    """
    Generate the Go model structural bias for a system
    of molecules. This processor class has two main
    functions: .contact_selector and .compute_bias.
    The .run_molecule function simply loops over all
    molecules in the system and calls the other two
    functions. The computed structural bias parameters
    are stored in `system.gmx_topology_params` and can
    be written out using the `vermouth.gmx.write_topology`
    function.

    Subclassing
    -----------
    In order to customize the Go-model structural bias
    it is recommended to subclass this function and
    overwrite the ``contact_selector`` method and/or
    the ``compute_bias`` method. This subclassed Processor
    then has to be added to the into the martinize2
    pipeline in place of the StructuralBiasWriter or as
    replacement in the GoPipeline.
    """
    def __init__(self,
                 contact_map,
                 cutoff_short,
                 cutoff_long,
                 go_eps,
                 res_dist,
                 moltype,
                 res_graph=None):
        """
        Initialize the Processor with arguments required
        to setup the Go model structural bias.

        Parameters
        ----------
        contact_map: list[(str, int, str, int)]
            list of contacts defined as by the chain
            identifier and residue index
        cutoff_short: float
            distances in nm smaller than this are ignored
        cutoff_long: float
            distances in nm larger than this are ignored
        go_eps: float
            epsilon value of the structural bias in
            kJ/mol
        res_dist: int
            if nodes are closer than res_dist along
            the residue graph they are ignored; this
            is similar to sequence distance but takes
            into account disulfide bridges for example
        moltype: str
            name of the molecule to treat
        res_graph: :class:`vermouth.molecule.Molecule`
            residue graph of the molecule; if None it
            gets generated automatically
        system: :class:`vermouth.system.System`
            the system
        magic_number: float
            magic number for Go contacts from the old
            GoVirt script.
        """
        self.contact_map = contact_map
        self.cutoff_short = cutoff_short
        self.cutoff_long = cutoff_long
        self.go_eps = go_eps
        self.res_dist = res_dist
        self.moltype = moltype
        # don't modify
        self.res_graph = None
        self.system = None
        self.__chain_id_to_resnode = {}
        self.conversion_factor = 2**(1/6)

    # do not overwrite when subclassing
    def _chain_id_to_resnode(self, chain, resid):
        """
        Return the node corresponding to the chain and
        resid. First time the function is run the dict
        is being created.

        Parameters
        ----------
        chain: str
            chain identifier
        resid: int
            residue index

        Returns
        -------
        dict
            a dict matching the chain,resid to the self.res_graph node
        """
        if self.__chain_id_to_resnode:
            return self.__chain_id_to_resnode[(chain, resid)]

        # for each residue collect the chain and residue in a dict
        # we use this later for identifying the residues from the
        # contact map
        for resnode in self.res_graph.nodes:
            chain_key = self.res_graph.nodes[resnode].get('chain', None)
            # in vermouth within a molecule all resid are unique
            # when merging multiple chains we store the old resid
            # the go model always references the input resid i.e.
            # the _old_resid
            resid_key = self.res_graph.nodes[resnode].get('_old_resid')
            self.__chain_id_to_resnode[(chain_key, resid_key)] = resnode

        return self.__chain_id_to_resnode[(chain, resid)]

    def contact_selector(self, molecule):
        """
        Select all contacts from the contact map
        that according to their distance and graph
        connectivity are elegible to form a Go
        bond and create exclusions between the
        backbone beads of those contacts.

        Parameters
        ----------
        molecule: :class:`vermouth.molecule.Molecule`

        Returns
        -------
        list[(collections.abc.Hashable, collections.abc.Hashable, float)]
            list of node keys and distance
        """
        # distance_matrix of elegible pairs as tuple(node, node, dist)
        contact_matrix = []
        # distance_matrix of elegible symmetrical pairs as tuple(node, node, dist)
        symmetrical_matrix = []
        # find all pairs of residues that are within bonded distance of
        # self.res_dist
        connected_pairs = dict(nx.all_pairs_shortest_path_length(self.res_graph,
                                                            cutoff=self.res_dist))
        for contact in self.contact_map:
            resIDA, chainA, resIDB, chainB = contact
            # identify the contact in the residue graph based on
            # chain ID and resid
            resA = self._chain_id_to_resnode(chainA, resIDA)
            resB = self._chain_id_to_resnode(chainB, resIDB)
            # make sure that both residues are not connected
            # note: contacts should be symmteric so we only
            # check against one
            if resB not in connected_pairs[resA]:
                # now we lookup the backbone nodes within the residue contact
                bb_node_A = next(filter_minimal(self.res_graph.nodes[resA]['graph'], select_backbone))
                bb_node_B = next(filter_minimal(self.res_graph.nodes[resB]['graph'], select_backbone))
                # compute the distance between bb-beads
                dist = np.linalg.norm(molecule.nodes[bb_node_A]['position'] -
                                      molecule.nodes[bb_node_B]['position'])
                # verify that the distance between BB-beads satisfies the
                # cut-off criteria
                if self.cutoff_long > dist > self.cutoff_short:
                    atype_a = next(get_go_type_from_attributes(self.res_graph.nodes[resA]['graph'],
                                                               _old_resid=resIDA,
                                                               chain=chainA,
                                                               prefix=self.moltype))
                    atype_b = next(get_go_type_from_attributes(self.res_graph.nodes[resB]['graph'],
                                                               _old_resid=resIDB,
                                                               chain=chainB,
                                                               prefix=self.moltype))
                    # Check if symmetric contact has already been processed before
                    # and if so, we append the contact to the final symmetric contact matrix
                    # and add the exclusions. Else, we add to the full valid contact_matrix
                    # and continue searching.
                    if (atype_b, atype_a, dist) in contact_matrix:
                        # generate backbone backbone exclusions
                        # perhaps one day can be it's own function
                        excl = Interaction(atoms=(bb_node_A, bb_node_B),
                                           parameters=[], meta={"group": "Go model exclusion"})
                        molecule.interactions['exclusions'].append(excl)
                        symmetrical_matrix.append((atype_a, atype_b, dist))
                    else:
                        contact_matrix.append((atype_a, atype_b, dist))

        return symmetrical_matrix

    def compute_go_interaction(self, contacts):
        """
        Compute the epsilon value given a distance between
        two nodes, figure out the atomtype name and store
        it in the systems attribute gmx_topology_params.

        Parameters
        ----------
        contacts: list[(str, str, float)]
            list of node-keys and their distance

        Returns
        ----------
        dict[frozenset(str, str): float]
            dict of interaction parameters indexed by atomtype
        """
        go_inters = {}
        for atype_a, atype_b, dist in contacts:
            sigma = dist / self.conversion_factor
            # find the go virtual-sites for this residue
            # probably can be done smarter but mehhhh
            contact_bias = NonbondParam(atoms=(atype_a, atype_b),
                                        sigma=sigma,
                                        epsilon=self.go_eps,
                                        meta={"comment": [f"go bond {dist}"]})
            self.system.gmx_topology_params["nonbond_params"].append(contact_bias)

    def run_molecule(self, molecule):
        self.res_graph = make_residue_graph(molecule)
        # compute the contacts; this also creates
        # the exclusions
        contacts = self.contact_selector(molecule)
        # compute the interaction parameters
        self.compute_go_interaction(contacts)
        return molecule

    def run_system(self, system):
        """
        Process `system`.

        Parameters
        ----------
        system: vermouth.system.System
            The system to process. Is modified in-place.
        """
        self.system = system
        super().run_system(system)



"""
I/O of topology parameters that are not molecules.
"""
import itertools
from collections import namedtuple
import textwrap
import vermouth
from vermouth.file_writer import deferred_open
from vermouth.citation_parser import citation_formatter
from ..log_helpers import StyleAdapter, get_logger
from .itp import _interaction_sorting_key

LOGGER = StyleAdapter(get_logger(__name__))

Atomtype = namedtuple('Atomtype', 'molecule node sigma epsilon meta')
NonbondParam = namedtuple('NonbondParam', 'atoms sigma epsilon meta')

def _group_by_conditionals(interactions):
    interactions_group_sorted = sorted(interactions,
        key=_interaction_sorting_key
        )
    interaction_grouped = itertools.groupby(
        interactions_group_sorted,
        key=_interaction_sorting_key
        )
    return interaction_grouped

def sigma_epsilon_to_C6_C12(sigma, epsilon):
    """
    Convert the LJ potential from sigma epsilon
    form to C6 C12 form.
    """
    C6 = 4*sigma*epsilon**6
    C12 = 4*sigma*epsilon**12
    return C6, C12

def write_atomtypes(system, itp_path, C6C12=False):
    """
    Writes the [atomtypes] directive to file.
    All atomtypes are defined in system.gmx_topology_params.
    Masses and further information are taken from the molecule
    directly.
    """
    conditional_keys = {True: '#ifdef', False: '#ifndef'}
    with deferred_open(itp_path, "w") as itp_file:
        itp_file.write("[ atomtypes ]\n")
        grouped_types = _group_by_conditionals(system.gmx_topology_params['atomtypes'])
        for (conditional, group), interactions_in_group in grouped_types:
            # conditionals are things like #ifdef; for more details on how this
            # works see the molecule_itp_writer
            if conditional:
                conditional_key = conditional_keys[conditional[1]]
                itp_file.write('{} {}\n'.format(conditional_key, conditional[0]))
            # groups are collections of interactions that are written bunched
            # together and indicated by a comment line
            if group:
                itp_file.write('; {}\n'.format(group))

            for atomtype in interactions_in_group:
                atype = atomtype.molecule.nodes[atomtype.node]['atype']
                charge = atomtype.molecule.nodes[atomtype.node]['charge']
                mass = atomtype.molecule.nodes[atomtype.node]['mass']

                if C6C12:
                    nb1, nb2 = sigma_epsilon_to_C6_C12(atomtype.sigma, atomtype.epsilon)
                else:
                    nb1, nb2 = atomtype.sigma, atomtype.epsilon

                if 'comment' in atomtype.meta:
                    comments = ";" + " ".join(atomtype.meta['comment'])
                else:
                    comments = ""
                itp_file.write(f"{atype} {mass} {charge} A {nb1:3.8F} {nb2:3.8F} {comments}\n")

def write_nonbond_params(system, itp_path, C6C12=False):
    """
    Writes the [nonbond_params] directive to file.
    All atomtypes are defined in system.gmx_topology_params.
    Masses and further information are taken from the molecule
    directly.
    """
    conditional_keys = {True: '#ifdef', False: '#ifndef'}
    with deferred_open(itp_path, "w") as itp_file:
        itp_file.write("[ nonbond_params ]\n")
        grouped_types = _group_by_conditionals(system.gmx_topology_params['nonbond_params'])
        for (conditional, group), interactions_in_group in grouped_types:
            if conditional:
                conditional_key = conditional_keys[conditional[1]]
                itp_file.write('{} {}\n'.format(conditional_key, conditional[0]))
            if group:
                itp_file.write('; {}\n'.format(group))


            for nb_params in interactions_in_group:
                if len(nb_params.atoms) == 2:
                    a1, a2 = nb_params.atoms
                # self interaction
                else:
                    a1 = nb_params.atoms[0]
                    a2 = nb_params.atoms[0]

                if C6C12:
                    nb1, nb2 = sigma_epsilon_to_C6_C12(nb_params.sigma, nb_params.epsilon)
                else:
                    nb1, nb2 = nb_params.sigma, nb_params.epsilon

                if nb_params.meta.get('comment'):
                    comments = ";" + " ".join(nb_params.meta['comment'])
                else:
                    comments = ""
                itp_file.write(f"{a1} {a2} 1 {nb1:3.8F} {nb2:3.8F} {comments}\n")

def write_gmx_topology(system,
                       top_path,
                       itp_paths={"nonbond_params": "extra_nbparams.itp",
                                  "atomtypes": "extra_atomtypes.itp"},
                       C6C12=False,
                       defines=(),
                       header=()):
    """
    Writes a Gromacs .top file for the specified system. Gromacs topology
    files are defined by directives for example `[ atomtypes ]`. However,
    Gromacs supports writing parts of the topology to so called .itp
    files which can be inculded into a toplevel topology file with the
    extension .top using #include statements. The topology writer will
    generate such a toplevel topology file where the different directives
    are written to seperate .itp files and included into the toplevel
    file.

    Parameters
    ----------
    system: vermouth.system.System
    top_path: pathlib.Path
        path for topology file
    itp_paths: dict[str, pathlib.Path]
        list of paths for writing the topology parameters
        like atomtypes with the key being the name of the
        directive.
    C6C12: bool
        write non-bonded interaction parameters using LJ
        C6C12 form
    defines: tuple(str)
        define statments to include in the topology
    header: tuple(str)
        any comment lines to include at the beginning
    """
    if not system.molecules:
        raise ValueError("No molecule in the system. Nothing to write.")

    include_string = ""
    # First we write the atomtypes directive
    if "atomtypes" in system.gmx_topology_params:
        _path = itp_paths['atomtypes']
        write_atomtypes(system, _path, C6C12)
        include_string += f'\n #include "{_path}"'
    # Next we write the nonbond_params directive
    if "nonbond_params" in system.gmx_topology_params:
        _path = itp_paths['nonbond_params']
        write_nonbond_params(system, _path, C6C12)
        include_string += f'\n #include "{_path}"\n'
    # Write the ITP files for the molecule types, and prepare writing the
    # [ molecules ] section of the top file.
    # * We write one ITP file for each different moltype in the system, the
    #   moltype being defined by the name provided under the "moltype" meta of
    #   the molecules. If more than one molecule share the same moltype, we use
    #   the first one to write the ITP file.
    moltype_written = set()
    # * We keep track of the length of the longer moltype name, to align the
    #   [ molecules ] section.
    max_name_length = 0
    # * We keep track of groups of successive molecules with the same moltypes.
    moltype_count = []  # items will be [moltype, number of molecules]

    # Iterate over groups of successive molecules with the same moltypes. We
    # shall *NOT* sort the molecules before hand, as groups of successive
    # molecules with the same moltype can be interupted by other moltypes, and
    # we want to reflect these interuptions in the [ molecules ] section of the
    # top file.
    molecule_groups = itertools.groupby(
        system.molecules, key=lambda x: x.meta["moltype"]
    )
    for moltype, molecules in molecule_groups:
        molecule = next(molecules)
        if moltype not in moltype_written:
            # A given moltype can appear more than once in the sequence of
            # molecules, without being uninterupted by other moltypes. Even in
            # that case, we want to write the ITP only once.
            with deferred_open("{}.itp".format(moltype), "w") as outfile:
                # here we format and merge all citations
                header[-1] = header[-1] + "\n"
                header.append("Please cite the following papers:")
                for citation in molecule.citations:
                    cite_string = citation_formatter(
                        molecule.force_field.citations[citation]
                    )
                    LOGGER.info("Please cite: " + cite_string)
                    header.append(cite_string)
                vermouth.gmx.itp.write_molecule_itp(molecule, outfile, header=header)
            this_moltype_len = len(molecule.meta["moltype"])
            if this_moltype_len > max_name_length:
                max_name_length = this_moltype_len
            moltype_written.add(moltype)
        # We already removed one element from the "molecules" generator, do not
        # forget to count it in the number of molecules in that group.
        moltype_count.append([moltype, 1 + len(list(molecules))])

    # Write the top file
    template = textwrap.dedent(
        """\
        {defines}
        #include "martini.itp"
        {includes}

        [ system ]
        Title of the system

        [ molecules ]
        {molecules}
    """
    )
    include_string = include_string + "\n".join(
        '#include "{}.itp"'.format(molecule_type) for molecule_type, _ in moltype_count
    )
    molecule_string = "\n".join(
        "{mtype:<{length}}    {num}".format(
            mtype=mtype, num=num, length=max_name_length
        )
        for mtype, num in moltype_count
    )
    define_string = "\n".join("#define {}".format(define) for define in defines)
    with deferred_open(str(top_path), "w") as outfile:
        outfile.write(
            textwrap.dedent(
                template.format(
                    includes=include_string,
                    molecules=molecule_string,
                    defines=define_string,
                )
            )
        )



# -*- coding: utf-8 -*-
# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the Go contact map reader.
"""
import pytest
from vermouth.rcsu.contact_map import read_go_map

@pytest.mark.parametrize('lines, contacts', 
        # two sets of contacts same chain
        (("""
         ID    I1  AA  C I(PDB)     I2  AA  C I(PDB)        DCA       CMs    rCSU   Count Model
         ============================================================================================
         R      1     1  LYS A    1        2  VAL A    2       3.8094     1 1 1 1    11     369    0

         R      6     1  LYS A    1       40  THR A   40       5.4657     1 1 1 1   222     238    0
         R     13     2  VAL A    2       37  ASN A   37       7.9443     0 1 0 1   -75      75    0
         R     15     2  VAL A    2       39  ASN A   39       4.2809     0 1 1 1   -13     121    0
         """,
         [(1, "A", 2, "A"), (1, "A", 40, "A"), (2, "A", 37, "A"), (2, "A", 39, "A")]
        ),
        # two sets of contacts same chain but skip two non-OV/rCSU
        ("""
         ID    I1  AA  C I(PDB)     I2  AA  C I(PDB)        DCA       CMs    rCSU   Count Model
         ============================================================================================
         R      1     1  LYS A    1        2  VAL A    2       3.8094     1 1 1 1    11     369    0

         R      6     1  LYS A    1       40  THR A   40       5.4657     0 1 1 0   222     238    0
         R     13     2  VAL A    2       37  ASN A   37       7.9443     0 1 0 0   -75      75    0
         R     15     2  VAL A    2       39  ASN A   39       4.2809     0 1 1 1   -13     121    0
         """,
         [(1, "A", 2, "A"), (2, "A", 39, "A")]
        ),
        # two sets of contacts different chains
        ("""
         ID    I1  AA  C I(PDB)     I2  AA  C I(PDB)        DCA       CMs    rCSU   Count Model
         ============================================================================================
         R      1     1  LYS A    1        2  VAL B    2       3.8094     1 1 1 1    11     369    0
         R      6     1  LYS A    1       40  THR B   40       5.4657     1 1 1 1   222     238    0
         R     13     2  VAL C    2       37  ASN D   37       7.9443     0 1 0 1   -75      75    0
         R     15     2  VAL C    2       39  ASN D   39       4.2809     0 1 1 1   -13     121    0
         """,
         [(1, "A", 2, "B"), (1, "A", 40, "B"), (2, "C", 37, "D"), (2, "C", 39, "D")]
        )))
def test_go_map(tmp_path, lines, contacts):
    # write the go contact map file
    with open(tmp_path / "go_file.txt", "w") as in_file:
        in_file.write(lines)

    # read go map
    contact_map = read_go_map(tmp_path / "go_file.txt")
    assert contact_map == contacts

def test_go_error(tmp_path):
    lines="""
          ID    I1  AA  C I(PDB)     I2  AA  C I(PDB)        DCA       CMs    rCSU   Count Model
          ============================================================================================
          No valid contacts in this file.
          """
    # write the go contact map file
    with open(tmp_path / "go_file.txt", "w") as in_file:
        in_file.write(lines)

    with pytest.raises(IOError):
        read_go_map(tmp_path / "go_file.txt")



# Copyright 2023 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for Go model processors.
"""
from vermouth.molecule import attributes_match

def get_go_type_from_attributes(molecule, prefix, **kwargs):
    """
    Find all nodes that satisfy a number of attributes specified
    as kwargs and have a specific atomtype prefix.

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
    prefix: str
        the atom-type prefix of the Go virtual side
    kwargs:
        any number of attributes

    Yields
    ------
    str
        the atom-type

    Raises
    ------
    KeyError
        If no node can be found that matches attributes
        and prefix an KeyError is raised.
    """
    for node in molecule.nodes:
        attrs = molecule.nodes[node]
        if attributes_match(attrs, kwargs) and attrs['atype'].startswith(prefix):
            yield attrs['atype']
    else:
        resid = kwargs['resid']
        chain = kwargs['chain']
        raise KeyError(f"Could not find GoVs with resid {resid} in chain {chain}.")

def _in_resid_region(resid, regions):
    """
    Check if resid falls in regions interval.

    Parameters
    ----------
    resid: int
        the resid of a molecule
    regions: list[tuple(int, int)]
        a list of the intervals

    Returns
    -------
    bool
    """
    for limits in regions:
        # perhaps someone gives them as reversed
        low, up = sorted(limits)
        if low <= resid <= up:
            return True
    return False

def _get_bead_size(atype):
    if atype.startswith("S"):
        bead_size = "small"
    elif atype.startswith("T"):
        bead_size = "tiny"
    else:
        bead_size = "regular"
    return bead_size



# Copyright 2023 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .processor import Processor
from ..graph_utils import make_residue_graph
from ..rcsu.go_utils import get_go_type_from_attributes, _get_bead_size, _in_resid_region
from ..gmx.topology import NonbondParam
import numpy as np

class ComputeWaterBias(Processor):
    """
    Processor which computes the water  bias for
    the Martini Go and Martini IDP model.

    The water bias strength is defined per secondary
    structure element in `water_bias` and assigned if
    `auto_bias` is set to True. Using the `idr_regions`
    argument the water_bias can be changed for
    intrinsically disordered regions (IDRs). The IDR
    bias superseeds the auto bias.

    This Processor updates the system.gmx_topology_params
    attribute.

    Subclassing
    -----------
    If the procedure by which to assign the water bias is
    to be changed this processor is best subclassed and the
    assign_residue_water_bias method overwritten.
    """

    def __init__(self,
                 auto_bias,
                 water_bias,
                 idr_regions):
        """


        Parameters
        ----------
        auto_bias: bool
            apply the automatic secondary structure
            dependent water biasing
        water_bias: dict[str, float]
            a dict of secondary structure codes and
            epsilon value for the water bias in kJ/mol
        idr_regions: list
            list of tuples of residue regions defining the IDRs
        prefix: str
            prefix of the Go virtual-site atomtypes
        system: vermouth.system.System
            the system of the molecules is used for
            storing the nonbonded parameters
        """
        self.water_bias = water_bias
        self.auto_bias = auto_bias
        self.idr_regions = idr_regions
        self.system = None

    def assign_residue_water_bias(self, molecule, res_graph):
        """
        Assign the residue water bias for all residues
        with a secondary structure element or that are
        defined by the region selector. Region selectors
        supercede the auto assignment.

        Parameters
        ----------
        molecule: :class:`vermouth.molecule.Molecule`
            the molecule
        res_graph: :class:`vermouth.molecule.Molecule`
            the residue graph of the molecule
        """
        for res_node in res_graph.nodes:
            resid = res_graph.nodes[res_node]['resid']
            _old_resid = res_graph.nodes[res_node]['_old_resid']
            chain = res_graph.nodes[res_node]['chain']
            resname = res_graph.nodes[res_node]['resname']

            if _in_resid_region(_old_resid, self.idr_regions):
                eps = self.water_bias.get('idr', 0.0)
                sec_struc = res_graph.nodes[res_node]['cgsecstruct']
            elif self.auto_bias:
                sec_struc = res_graph.nodes[res_node]['cgsecstruct']
                eps = self.water_bias.get(sec_struc, 0.0)
            else:
                continue

            if abs(eps) <= 1e-12:
                continue

            vs_go_node = next(get_go_type_from_attributes(res_graph.nodes[res_node]['graph'],
                                                          resid=resid,
                                                          chain=chain,
                                                          prefix=molecule.meta.get('moltype')))

            # what is the blocks bb-type
            bb_type = molecule.force_field.blocks[resname].nodes['BB']['atype']
            size = _get_bead_size(bb_type)
            # bead sizes are defined in the force-field file as
            # regular, small and tiny
            sigma = float(molecule.force_field.variables[size])
            # update interaction parameters
            atoms = (molecule.force_field.variables['water_type'], vs_go_node)
            water_bias = NonbondParam(atoms=atoms,
                                      sigma=sigma,
                                      epsilon=eps,
                                      meta={"comment": ["water bias", sec_struc]})
            self.system.gmx_topology_params["nonbond_params"].append(water_bias)

    def remove_cross_nb_interactions(self, molecule, res_graph):
        """
        Remove Go bonds between folded and disordered regions of a molecule

        Parameters
        ----------
        molecule: :class:`vermouth.molecule.Molecule`
            the molecule
        res_graph: :class:`vermouth.molecule.Molecule`
            the residue graph of the molecule
        """
        #list of all the Go pairs in the molecule
        all_go_pairs = np.array([list(i.atoms) for i in self.system.gmx_topology_params["nonbond_params"] if 'W' not in list(i.atoms)])
        # list to record which items we don't want. cross = go potential between folded and disordered domain.
        all_cross_pairs = []

        for res_node in res_graph.nodes:
            resid = res_graph.nodes[res_node]['resid']
            _old_resid = res_graph.nodes[res_node]['_old_resid']
            chain = res_graph.nodes[res_node]['chain']

            if _in_resid_region(_old_resid, self.idr_regions):
                vs_go_node = next(get_go_type_from_attributes(res_graph.nodes[res_node]['graph'],
                                                              resid=resid,
                                                              chain=chain,
                                                              prefix=molecule.meta.get('moltype')))
                all_cross_pairs.append(np.where(all_go_pairs == vs_go_node)[0]) #just need the first one

        # make sure we only have one entry in case a site has more than one interaction
        all_cross_pairs = np.unique([x for xs in all_cross_pairs for x in xs])
        # delete the folded-disordered Go interactions from the list going backwards.
        # otherwise list order gets messed up.
        for i in reversed(all_cross_pairs):
            del self.system.gmx_topology_params["nonbond_params"][i]

    def run_molecule(self, molecule):
        """
        Assign water bias for a single molecule
        """
        if not self.system:
            raise IOError('This processor requires a system.')

        if not molecule.meta.get('moltype'):
            raise ValueError('The molecule does not have a moltype name.')

        if hasattr(molecule, 'res_graph'):
            res_graph = molecule.res_graph
        else:
            res_graph = make_residue_graph(molecule)

        self.assign_residue_water_bias(molecule, res_graph)
        self.remove_cross_nb_interactions(molecule, res_graph)

        return molecule

    def run_system(self, system):
        """
        Assign the water bias of the Go model to file. Biasing
        is always molecule specific i.e. no two different
        vermouth molecules can have the same bias.

        Parameters
        ----------
        system: :class:`vermouth.system.System`
        """
        if not (self.idr_regions or self.auto_bias):
            return system
        self.system = system
        super().run_system(system)



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the writing of TOP file.
"""
import os
import textwrap
import pytest
import vermouth
from vermouth.file_writer import DeferredFileWriter
from vermouth.gmx.topology import (sigma_epsilon_to_C6_C12,
                                   write_atomtypes,
                                   write_nonbond_params,
                                   write_gmx_topology,
                                   Atomtype,
                                   NonbondParam)
from vermouth.tests.gmx.test_itp import dummy_molecule

@pytest.mark.parametrize('atomtypes, expected, C6C12',
        # simple atomtype
        #'Atomtype', 'molecule node sigma epsilon meta
        ((
        [{"node": 0, "sigma": 0.43, "epsilon": 2.3, "meta": {}}],
        ["[ atomtypes ]\n", 
         "A 72 0 A 0.43000000 2.30000000 \n"], False),
        (
        [{"node": 0, "sigma": 0.43, "epsilon": 2.3, "meta": {}}],
        ["[ atomtypes ]\n", 
         "A 72 0 A 254.62172908 37693.15402307 \n"], True),
        (
        [{"node":0, "sigma": 0.43, "epsilon": 2.3, "meta": {}},
         {"node":1, "sigma": 0.47, "epsilon": 2.5, "meta": {}}],
        ["[ atomtypes ]\n", 
         "A 72 0 A 0.43000000 2.30000000 \n",
         "B 72 0 A 0.47000000 2.50000000 \n"], False),
        (
        [{"node": 0, "sigma": 0.43, "epsilon": 2.3, "meta": {"comment": ["comment"]}}],
        ["[ atomtypes ]\n", 
         "A 72 0 A 0.43000000 2.30000000 ;comment\n"], False),
        (
        [{"node":0, "sigma": 0.43, "epsilon": 2.3, "meta": {"group": "g1"}},
         {"node":1, "sigma": 0.44, "epsilon": 3.3, "meta": {}},
         {"node":2, "sigma": 0.47, "epsilon": 2.5, "meta": {"group": "g1"}}],
        ['[ atomtypes ]\n', 
         'B 72 0 A 0.44000000 3.30000000 \n', 
         '; g1\n', 
         'A 72 0 A 0.43000000 2.30000000 \n', 
         'C 72 0 A 0.47000000 2.50000000 \n'], False
        ),
        (
        [{"node":0, "sigma": 0.43, "epsilon": 2.3, "meta": {"ifdef": "g1"}},
         {"node":1, "sigma": 0.44, "epsilon": 3.3, "meta": {}},
         {"node":2, "sigma": 0.47, "epsilon": 2.5, "meta": {"ifdef": "g1"}}],
        ['[ atomtypes ]\n', 
         'B 72 0 A 0.44000000 3.30000000 \n', 
         '#ifdef g1\n', 
         'A 72 0 A 0.43000000 2.30000000 \n', 
         'C 72 0 A 0.47000000 2.50000000 \n',
         '#endif'], False
        ),
        ))
def test_atomtypes(tmp_path, dummy_molecule, atomtypes, expected, C6C12):
    """
    Test that the atomtypes directive is properly written.
    """
    dummy_sys = vermouth.system.System("")
    for atype in atomtypes:
        # somewhat hacky workaround because pytest doesn't
        # allow the use of fixtures in parametrize
        atype['molecule'] = dummy_molecule
        dummy_sys.gmx_topology_params['atomtypes'].append(Atomtype(**atype))

    outpath = tmp_path / 'out.itp'
    write_atomtypes(dummy_sys, outpath, C6C12=C6C12)
    DeferredFileWriter().write()

    with open(str(outpath)) as infile:
        for line, ref_line in zip(infile, expected):
            assert line == ref_line


@pytest.mark.parametrize('nbparams, expected, C6C12',
        # simple atomtype
        #'Atomtype', 'molecule node sigma epsilon meta
        ((
        [{"atoms": ("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {}}],
        ["[ nonbond_params ]\n", 
         "A B 1 0.43000000 2.30000000 \n"], False),
        (
        [{"atoms": ("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {}}],
        ["[ nonbond_params ]\n", 
         "A B 1 254.62172908 37693.15402307 \n"], True),
        (
        [{"atoms": ("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {}},
         {"atoms": ("A", "C"), "sigma": 0.47, "epsilon": 2.5, "meta": {}}],
        ["[ nonbond_params ]\n", 
         "A B 1 0.43000000 2.30000000 \n",
         "A C 1 0.47000000 2.50000000 \n"], False),
        # self interaction
        (
        [{"atoms": ("A",), "sigma": 0.43, "epsilon": 2.3, "meta": {}},
         {"atoms": ("A", "C"), "sigma": 0.47, "epsilon": 2.5, "meta": {}}],
        ["[ nonbond_params ]\n", 
         "A A 1 0.43000000 2.30000000 \n",
         "A C 1 0.47000000 2.50000000 \n"], False),
        (
        [{"atoms": ("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {"comment": ["comment"]}}],
        ["[ nonbond_params ]\n", 
         "A B 1 0.43000000 2.30000000 ;comment\n"], False),
        (
        [{"atoms":("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {"group": "g1"}},
         {"atoms":("B", "C"), "sigma": 0.44, "epsilon": 3.3, "meta": {}},
         {"atoms":("A", "C"), "sigma": 0.47, "epsilon": 2.5, "meta": {"group": "g1"}}],
        ['[ nonbond_params ]\n', 
         'B C 1 0.44000000 3.30000000 \n', 
         '; g1\n', 
         'A B 1 0.43000000 2.30000000 \n', 
         'A C 1 0.47000000 2.50000000 \n'], False
        ),
        (
        [{"atoms":("A", "B"), "sigma": 0.43, "epsilon": 2.3, "meta": {"ifdef": "g1"}},
         {"atoms":("B", "C"), "sigma": 0.44, "epsilon": 3.3, "meta": {}},
         {"atoms":("A", "C"), "sigma": 0.47, "epsilon": 2.5, "meta": {"ifdef": "g1"}}],
        ['[ nonbond_params ]\n', 
         'B C 1 0.44000000 3.30000000 \n', 
         '#ifdef g1\n', 
         'A B 1 0.43000000 2.30000000 \n', 
         'A C 1 0.47000000 2.50000000 \n',
         '#endif'], False
        ),
        ))
def test_nonbond_params(tmp_path, nbparams, expected, C6C12):
    """
    Test that the atomtypes directive is properly written.
    """
    dummy_sys = vermouth.system.System("")
    for nbparam in nbparams:
        dummy_sys.gmx_topology_params['nonbond_params'].append(NonbondParam(**nbparam))

    outpath = tmp_path / 'out.itp'
    write_nonbond_params(dummy_sys, outpath, C6C12=C6C12)
    DeferredFileWriter().write()

    with open(str(outpath)) as infile:
        for line, ref_line in zip(infile, expected):
            assert line == ref_line

def test_toplevel_topology(tmp_path, dummy_molecule):
    """
    Make sure the toplevel topology file
    is correctly written. Note that the individual
    molecule itp files are seperately checked in
    gmx itp module or in the previous tests.
    """
    os.chdir(tmp_path)
    system = vermouth.System()
    system.molecules.append(dummy_molecule)
    dummy_molecule.meta['moltype'] = "molecule_0"
    # "node": 0, "sigma": 0.43, "epsilon": 2.3, "meta": {}}
    system.gmx_topology_params['atomtypes'].append(Atomtype(node=0,
                                                   molecule=dummy_molecule,
                                                   sigma=0.43,
                                                   meta={},
                                                   epsilon=2.3))
    system.gmx_topology_params['nonbond_params'].append(NonbondParam(atoms=("A", "B"),
                                                                     sigma=0.43,
                                                                     epsilon=2.3,
                                                                     meta={}))
    outpath = tmp_path / 'out.itp'
    atompath = tmp_path / 'atomtypes.itp'
    nbpath = tmp_path / 'nonbond_params.itp'
    write_gmx_topology(system,
                       outpath,
                       header=['first header comment', 'second header comment'],
                       defines=('random', ),
                       itp_paths={"atomtypes": atompath, "nonbond_params": nbpath},
                       # at this level C6C12 doesn't matter; it gets
                       # checked in previous texts
                       C6C12=False)
    DeferredFileWriter().write()

    reference =f"""#define random
#include "martini.itp"

#include "{tmp_path}/atomtypes.itp"
#include "{tmp_path}/nonbond_params.itp"
#include "molecule_0.itp"

[ system ]
Title of the system

[ molecules ]
molecule_0    1
"""
    ref_lines = textwrap.dedent(reference).splitlines()
    with open(str(outpath)) as infile:
        for line, ref_line in zip(infile, ref_lines):
            assert line.strip() == ref_line



# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test for the water bias processor.
"""
import pytest
import vermouth
from vermouth.rcsu.go_vs_includes import VirtualSiteCreator
from vermouth.processors.water_bias import ComputeWaterBias
from vermouth.tests.helper_functions import create_sys_all_attrs, test_molecule
from vermouth.gmx.topology import NonbondParam

@pytest.mark.parametrize('secstruc, water_bias, idr_regions, expected',
        ((
        # only auto-bias single sec struct
        {1: "H", 2: "H", 3: "H", 4: "H"},
        {"H": 2.1},
        [],
        {0: "H", 3: "H", 5: "H", 6: "H"}
        ),
        # test skip unassigned node
        ({1: None, 2: "H", 3: "H", 4: "H"},
        {"H": 2.1},
        [],
        {3: "H", 5: "H", 6: "H"}
        ),
        # only auto-bias two sec struct
        ({1: "H", 2: "H", 3: "C", 4: "C"},
        {"H": 2.1, "C": 3.1},
        [],
        {0: "H", 3: "H", 5: "C", 6: "C"}
        ),
        # only idp bias
        ({1: "H", 2: "H", 3: "C", 4: "C"},
        {"idr": 2.1},
        [(2, 3)],
        {3: "idr", 5: "idr"}
        ),
        # idp and sec struc bias
        ({1: "H", 2: "H", 3: "C", 4: "C"},
        {"idr": 1.1, "C": 3.1, "H": 2.1},
        [(2, 3)],
        {0: "H", 3: "idr", 5: "idr", 6: "C"}
        )
        ))
def test_assign_residue_water_bias(test_molecule, 
                                   secstruc,
                                   water_bias,
                                   idr_regions,
                                   expected):
    # bead sizes
    sizes = {0: 0.47, 3: 0.41, 5: 0.38, 6: 0.47}
    # the molecule atomtypes
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1", 
              6: "P1", 7: "SN3a", 8: "SP4"}
    # the molecule resnames
    resnames = {0: "A", 1: "A", 2: "A",
                3: "B", 4: "B",
                5: "C", 
                6: "D", 7: "D", 8: "D"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    # generate the virtual sites
    VirtualSiteCreator().run_system(system)

    processor = ComputeWaterBias(water_bias=water_bias,
                                 auto_bias=True,
                                 idr_regions=idr_regions)
    processor.run_system(system)
    for nb_params in system.gmx_topology_params['nonbond_params']:
        assert nb_params.atoms[0] == "W"
        vs_node_atype = nb_params.atoms[1]
        for atom in system.molecules[0].atoms:
            if atom[1]['atype'] == vs_node_atype:
                vs_node = atom[0]
        for vs in system.molecules[0].interactions['virtual_sitesn']:
            if vs.atoms[0] == vs_node:
                bb_node = vs.atoms[1]
                water_eps = water_bias[expected[bb_node]]
                water_sig = sizes[bb_node]
                assert water_eps == nb_params.epsilon
                assert water_sig == nb_params.sigma

@pytest.mark.parametrize('idr_regions, go_bonds, expected',
         (
            (#no idrs to remove
             [],
             [[1, 3], [2, 4]],
             [[1, 3], [2, 4]]
            ),
            (#central idr to remove bonds from
             [(2, 3)],
             [[1, 3], [2, 4], [1, 4]],
             [[1, 4]]
            ),
            (#remove all bonds within a region
            [(1, 4)],
            [[1, 2], [2, 3], [3, 4], [1, 4]],
            []
            )
          ))
def test_cross_go_bond_removal(test_molecule,
                               idr_regions,
                               go_bonds,
                               expected):
    # bead sizes
    sizes = {0: 0.47, 3: 0.41, 5: 0.38, 6: 0.47}
    # the molecule atomtypes
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}
    # the molecule resnames
    resnames = {0: "A", 1: "A", 2: "A",
                3: "B", 4: "B",
                5: "C",
                6: "D", 7: "D", 8: "D"}
    secstruc = {1: "H", 2: "H", 3: "H", 4: "H"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    # generate the virtual sites
    VirtualSiteCreator().run_system(system)

    #add the go bonds between residues
    for go_bond in go_bonds:
        atypes = []
        for index in go_bond:
            for atom in [i for i in system.molecules[0].atoms]:
                if (atom[1]['resid'] == index) and (atom[1]['atomname'] == 'CA'):
                    atypes.append(atom[1]['atype'])
        contact_bias = NonbondParam(atoms=(atypes[0], atypes[1]),
                                sigma=0.5, #these values don't matter
                                epsilon=0.5,
                                meta={"comment": ["go bond"]})
        system.gmx_topology_params["nonbond_params"].append(contact_bias)

    #apply water bias and remove folded-disordered domain bonds
    processor = ComputeWaterBias(water_bias={"idr": 1.1, "C": 3.1, "H": 2.1}, # doesn't matter
                                 auto_bias=True,
                                 idr_regions=idr_regions)
    processor.run_system(system)

    #find the go bonds which remain after removal, and don't involve water
    remaining = [list(i.atoms) for i in system.gmx_topology_params["nonbond_params"] if 'W' not in list(i.atoms)]

    assert len(remaining) == len(expected)
    for i in expected:
        expected_names = [f"molecule_0_{j}" for j in i]

        assert expected_names in remaining

def test_no_moltype_error(test_molecule):
    """
    Test that various high level errors are
    properly raised.
    """
    # set up processor
    processor = ComputeWaterBias(water_bias={"C": 3.1},
                                 auto_bias=True,
                                 idr_regions=None)
    # no moltype set
    system = vermouth.System()
    system.add_molecule(test_molecule)
    with pytest.raises(ValueError):
        processor.run_system(system)

def test_no_system_error(test_molecule):
    """
    Test that various high level errors are
    properly raised.
    """
    # set up processor
    processor = ComputeWaterBias(water_bias={"C": 3.1},
                                 auto_bias=True,
                                 idr_regions=None)
    test_molecule.meta['moltype'] = "random"
    # no system
    with pytest.raises(IOError):
        processor.run_molecule(test_molecule)

def test_clean_return(test_molecule):
    # set up processor
    processor = ComputeWaterBias(water_bias={"C": 3.1},
                                 auto_bias=None,
                                 idr_regions=None)
    test_molecule.meta['moltype'] = "random"
    system = vermouth.System()
    system.add_molecule(test_molecule)
    assert processor.run_system(system) == system



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains unittests for vermouth.processors.merge_chains.
"""

import networkx as nx
import pytest
from vermouth.system import System
from vermouth.molecule import Molecule
from vermouth.forcefield import ForceField
from vermouth.processors.merge_chains import (
    MergeChains
)
from vermouth.tests.datafiles import (
    FF_UNIVERSAL_TEST,
)

@pytest.mark.parametrize('node_data, edge_data, merger, expected', [
    (
            [
                {'chain': 'A', 'resname': 'ALA', 'resid': 1},
                {'chain': 'A', 'resname': 'ALA', 'resid': 2},
                {'chain': 'A', 'resname': 'ALA', 'resid': 3},
                {'chain': 'B', 'resname': 'ALA', 'resid': 1},
                {'chain': 'B', 'resname': 'ALA', 'resid': 2},
                {'chain': 'B', 'resname': 'ALA', 'resid': 3}
            ],
            [(0, 1), (1, 2), (3, 4), (4, 5)],
            {"chains": ["A", "B"], "all_chains": False},
            False
    ),
    (
            [
                {'chain': 'A', 'resname': 'ALA', 'resid': 1},
                {'chain': 'A', 'resname': 'ALA', 'resid': 2},
                {'chain': 'A', 'resname': 'ALA', 'resid': 3},
                {'chain': 'B', 'resname': 'ALA', 'resid': 1},
                {'chain': 'B', 'resname': 'ALA', 'resid': 2},
                {'chain': 'B', 'resname': 'ALA', 'resid': 3}
            ],
            [(0, 1), (1, 2), (3, 4), (4, 5)],
            {"chains": [], "all_chains": True},
            False
    ),
    (
            [
                {'chain': 'A', 'resname': 'ALA', 'resid': 1},
                {'chain': 'A', 'resname': 'ALA', 'resid': 2},
                {'chain': 'A', 'resname': 'ALA', 'resid': 3},
                {'chain': None, 'resname': 'ALA', 'resid': 1},
                {'chain': None, 'resname': 'ALA', 'resid': 2},
                {'chain': None, 'resname': 'ALA', 'resid': 3}
            ],
            [(0, 1), (1, 2), (3, 4), (4, 5)],
            {"chains": [], "all_chains": True},
            True
    ),

])
def test_merge(caplog, node_data, edge_data, merger, expected):
    """
    Tests that the merging works as expected.
    """
    system = System(force_field=ForceField(FF_UNIVERSAL_TEST))
    mol = Molecule(force_field=system.force_field)
    mol.add_nodes_from(enumerate(node_data))
    mol.add_edges_from(edge_data)

    mols = nx.connected_components(mol)
    for nodes in mols:
        system.add_molecule(mol.subgraph(nodes))

    processor = MergeChains(**merger)
    caplog.clear()
    processor.run_system(system)

    if expected:
        assert any(rec.levelname == 'WARNING' for rec in caplog.records)
    else:
        assert caplog.records == []

def test_too_many_args():
    """
    Tests that error is raised when too many arguments are given.
    """
    node_data = [
                {'chain': 'A', 'resname': 'ALA', 'resid': 1},
                {'chain': 'A', 'resname': 'ALA', 'resid': 2},
                {'chain': 'A', 'resname': 'ALA', 'resid': 3},
                {'chain': 'B', 'resname': 'ALA', 'resid': 1},
                {'chain': 'B', 'resname': 'ALA', 'resid': 2},
                {'chain': 'B', 'resname': 'ALA', 'resid': 3}
                ]
    edge_data = [(0, 1), (1, 2), (3, 4), (4, 5)]

    system = System(force_field=ForceField(FF_UNIVERSAL_TEST))
    mol = Molecule(force_field=system.force_field)
    mol.add_nodes_from(enumerate(node_data))
    mol.add_edges_from(edge_data)

    mols = nx.connected_components(mol)
    for nodes in mols:
        system.add_molecule(mol.subgraph(nodes))

    merger = {"chains": ["A", "B"], "all_chains": True}

    processor = MergeChains(**merger)

    with pytest.raises(ValueError):
        processor.run_system(system)




# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains helper functions for tests.
"""
import operator
import os
import pytest
import numpy as np
import networkx as nx
import networkx.algorithms.isomorphism as iso
import vermouth
from vermouth.system import System
from vermouth.forcefield import ForceField

def make_into_set(iter_of_dict):
    """
    Convenience function that turns an iterator of dicts into a set of
    frozenset of the dict items.
    """
    return set(frozenset(dict_.items()) for dict_ in iter_of_dict)


def equal_graphs(g1, g2,
                 node_attrs=('resid', 'resname', 'atomname', 'chain', 'charge_group', 'atype'),
                 edge_attrs=()):
    """
    Parameters
    ----------
    g1: networkx.Graph
    g2: networkx.Graph
    node_attrs: collections.abc.Iterable or None
        Node attributes to consider. If `None`, the node attribute dicts must
        be equal.
    edge_attrs: collections.abc.Iterable or None
        Edge attributes to consider. If `None`, the edge attribute dicts must
        be equal.

    Returns
    -------
    bool
        True if `g1` and `g2` are isomorphic, False otherwise.
    """
    if node_attrs is None:
        node_equal = operator.eq
    else:
        node_equal = iso.categorical_node_match(node_attrs, [''] * len(node_attrs))
    if edge_attrs is None:
        edge_equal = operator.eq
    else:
        edge_equal = iso.categorical_node_match(edge_attrs, [''] * len(edge_attrs))
    matcher = iso.GraphMatcher(g1, g2, node_match=node_equal, edge_match=edge_equal)
    return matcher.is_isomorphic()


def find_in_path(names=('martinize2', 'martinize2.py')):
    """
    Finds and returns the location of one of `names` in PATH, and returns the
    first match.

    Parameters
    ----------
    names: collections.abc.Sequence
        Names to look for in PATH.

    Returns
    -------
    os.PathLike or None
    """
    for folder in os.getenv("PATH", '').split(os.pathsep):
        for name in names:
            fullpath = os.path.join(folder, name)
            if os.path.isfile(fullpath):
                return fullpath
          
def create_sys_all_attrs(molecule, moltype, secstruc, defaults, attrs):
    """
    Generate a test system from a molecule
    with all attributes set and blocks in
    force-field.

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
    moltype: str
        sets meta['moltype']
    secstruc: dict[int, str]
        secondary structure attributes
        as resid str pairs
    defaults: dict
        dict of attribute default value
    attrs: dict
        dict of attribute name and dict
        of node value pairs

    Returns
    -------
    :class:`vermouth.system.System`
    """
    # set mol meta
    molecule.meta['moltype'] = moltype
    # assign default node attributes
    for attr, value in defaults.items():
        nx.set_node_attributes(molecule, value, attr)

    # assign node attributes
    for attr, values in attrs.items():
        nx.set_node_attributes(molecule, values, attr)

    # assign resids
    resids = nx.get_node_attributes(molecule, "resid")
    nx.set_node_attributes(molecule, resids, "_old_resid")
   
    # make the proper force-field
    ff = ForceField("test")
    ff.variables['water_type'] = "W"
    ff.variables['regular'] = 0.47
    ff.variables['small'] = 0.41
    ff.variables['tiny'] = 0.38

    res_graph = vermouth.graph_utils.make_residue_graph(molecule)
    for node in res_graph.nodes:
        mol_nodes = res_graph.nodes[node]['graph'].nodes
        block = vermouth.molecule.Block()
        resname = res_graph.nodes[node]['resname']
        resid = res_graph.nodes[node]['resid']
        # assign secondary structure
        for mol_node in mol_nodes:
            molecule.nodes[mol_node]['cgsecstruct'] = secstruc[resid]
            block.add_node(molecule.nodes[mol_node]['atomname'],
                           atype=molecule.nodes[mol_node]['atype'])

        ff.blocks[resname] = block


    # create the system
    molecule._force_field = ff
    system = System()
    system.molecules.append(molecule)
    return system



def parse_gofiles(file, atomtypes=False):
    '''
    Parser of go_nbparams.itp & go_atomtypes.itp files into an easy to assert dictionary.
    '''
    with open(file) as my_file:
        next(my_file)  # Skip header
        vals = {}
        for line in my_file:
            line = line.split()
            if atomtypes:
                # Key is atomname str, value is atomdef str
                vals[line[0]] = ' '.join(line[1:])
            else:
                # Key is tuple with nb pair, value is tuple with nb sigma and eps
                tup = tuple(sorted((line[0], line[1])))
                vals[tup] = tuple((float(line[3]), float(line[4])))
    return vals

@pytest.fixture
def test_molecule(scope='function'):
    """
    Molecule with the following connectivity and atom-naming:

    SC2:   2           8
           |           |
    SC1:   1   4       7
           |   |       |
    BB:    0 - 3 - 5 - 6
           -------------
    resid: 1   2   3   4  column wise
    """

    force_field = vermouth.forcefield.ForceField("test")
    molecule = vermouth.molecule.Molecule(force_field=force_field)
    molecule.meta['test'] = True
    # The node keys should not be in a sorted order as it would mask any issue
    # due to the keys being accidentally sorted.
    molecule.add_node(2, atomname='SC2',
                      position=np.array([0., 1.0, 0.0]), resid=1)
    molecule.add_node(0, atomname='BB',
                      position=np.array([0., 0., 0.]), resid=1)
    molecule.add_node(1, atomname='SC1',
                      position=np.array([0., 0.5, 0.0]), resid=1)

    molecule.add_node(3, atomname='BB', position=np.array(
        [0.5, 0.0, 0.0]), resid=2)
    molecule.add_node(4, atomname='SC1', position=np.array(
        [0.5, 0.5, 0.0]), resid=2)

    molecule.add_node(5, atomname='BB', position=np.array(
        [1.0, 0.0, 0.0]), resid=3)

    molecule.add_node(6, atomname='BB', position=np.array(
        [1.5, 0.0, 0.0]), resid=4)
    molecule.add_node(7, atomname='SC1', position=np.array(
        [1.5, 0.5, 0.0]), resid=4)
    molecule.add_node(8, atomname='SC2', position=np.array(
        [1.5, 1.0, 0.0]), resid=4)

    molecule.add_edge(0, 1)
    molecule.add_edge(0, 2)
    molecule.add_edge(0, 3)
    molecule.add_edge(3, 4)
    molecule.add_edge(3, 5)
    molecule.add_edge(5, 6)
    molecule.add_edge(6, 7)
    molecule.add_edge(7, 8)

    return molecule




# Copyright 2023 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrapper of Processors defining the GoPipline.
"""
import networkx as nx
import inspect
import vermouth
from ..processors.processor import Processor
from .go_vs_includes import VirtualSiteCreator
from .go_structure_bias import ComputeStructuralGoBias
from ..processors import SetMoleculeMeta

class GoProcessorPipline(Processor):
    """
    Wrapping all processors for the go model.
    """
    def __init__(self, processor_list):
        self.processor_list = processor_list
        self.kwargs = {}

    def prepare_run(self, system, moltype):
        """
        Things to do before running the pipeline.
        """
        # merge all molecules in the system
        # this will eventually become deprecated
        # with the proper Go-model for multimers
        vermouth.MergeAllMolecules().run_system(system)
        molecule = system.molecules[0]
        molecule.meta['moltype'] = moltype

    def run_system(self, system, **kwargs):
        self.kwargs = kwargs
        self.prepare_run(system, moltype=kwargs['moltype'])
        for processor in self.processor_list:
            process_args = inspect.getfullargspec(processor).args
            process_args_values = {arg:self.kwargs[arg] for arg in kwargs.keys() if arg in process_args}
            processor(**process_args_values).run_system(system)
        return system

GoPipeline = GoProcessorPipline([SetMoleculeMeta,
                                 VirtualSiteCreator,
                                 ComputeStructuralGoBias])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides processors that can add and remove IDR specific bonds
"""

from .processor import Processor
from ..rcsu.go_utils import _in_resid_region
from ..log_helpers import StyleAdapter, get_logger
LOGGER = StyleAdapter(get_logger(__name__))


def annotate_disorder(molecule, id_regions, annotation="cgidr"):
    """
    Annotate the disordered regions of the molecule

    molecule: :class:`vermouth.molecule.Molecule`
            the molecule
    idr_regions: list
        list of tuples defining the resids of the idrs in the molecule
    annotation: str
        name of the annotation in the node
    """

    for key, node in molecule.nodes.items():
        _old_resid = node['_old_resid']
        if _in_resid_region(_old_resid, id_regions):
            molecule.nodes[key][annotation] = True

class AnnotateIDRs(Processor):
    """
    Processor to annotate intrinsically disordered regions of a molecule.

    This processor is designed primarily for the work described in the reference
    M3_GO, but is generally applicable for such circumstances where extra
    addition/removals are necessary.

    """

    def __init__(self, id_regions=None):
        """
        Parameters
        ----------
        id_regions:
            regions defining the IDRs
        """
        self.id_regions = id_regions

    def run_molecule(self, molecule):
        """
        Assign disordered regions for a single molecule
        """

        annotate_disorder(molecule, self.id_regions)

        return molecule

    def run_system(self, system):
        """
        Assign the water bias of the Go model to file. Biasing
        is always molecule specific i.e. no two different
        vermouth molecules can have the same bias.

        Parameters
        ----------
        system: :class:`vermouth.system.System`
        """
        if not self.id_regions:
            return system
        LOGGER.info("Annotating disordered regions", type="step")
        super().run_system(system)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from vermouth.processors import do_links, DoLinks
from vermouth.molecule import Molecule, Link, Modification, Choice
from vermouth.processors.do_links import _atoms_match
import vermouth.forcefield


@pytest.mark.parametrize(
    "orders, resids, answer", (
        # 0, 0
        ((0, 0), (4, 4), True),
        ((0, 0), (3, 4), False),
        # 0, n
        ((0, -3), (6, 3), True),
        ((0, -3), (6, 6), False),
        ((0, -3), (6, -2), False),
        ((0, -3), (6, 9), False),
        ((0, 3), (6, 9), True),
        ((0, 3), (6, 6), False),
        ((0, 3), (6, 3), False),
        # 0, ><
        ((0, '>'), (5, 8), True),
        ((0, '>'), (5, 5), False),
        ((0, '>'), (5, 3), False),
        ((0, '>>'), (5, 8), True),
        ((0, '>>'), (5, 5), False),
        ((0, '>>'), (5, 3), False),
        ((0, '<'), (5, 3), True),
        ((0, '<'), (5, 5), False),
        ((0, '<'), (5, 8), False),
        ((0, '<<'), (5, 3), True),
        ((0, '<<'), (5, 5), False),
        ((0, '<<'), (5, 8), False),
        # 0, *
        ((0, '*'), (4, 1), True),
        ((0, '*'), (4, 8), True),
        ((0, '*'), (4, 4), False),
        ((0, '**'), (4, 1), True),
        ((0, '**'), (4, 8), True),
        ((0, '**'), (4, 4), False),
        # n, 0
        ((4, 0), (6, 2), True),
        ((4, 0), (6, 8), False),
        ((-4, 0), (2, 6), True),
        ((-4, 0), (4, 6), False),
        # n, n
        ((5, 5), (7, 7), True),
        ((5, 5), (7, 8), False),
        ((5, 7), (2, 4), True),
        ((5, 7), (2, 5), False),
        ((5, 7), (2, 1), False),
        ((7, 5), (4, 2), True),
        ((7, 5), (2, 5), False),
        ((7, 5), (2, 1), False),
        ((-2, -3), (5, 4), True),
        ((-2, -3), (5, 8), False),
        # n, ><  no comparison
        ((3, '>'), (8, 9), True),
        ((3, '>'), (8, 8), True),
        ((3, '>'), (8, 5), True),
        ((3, '>>'), (8, 9), True),
        ((3, '>>'), (8, 8), True),
        ((3, '>>'), (8, 5), True),
        ((3, '<'), (8, 6), True),
        ((3, '<'), (8, 8), True),
        ((3, '<'), (8, 9), True),
        ((3, '<<'), (8, 6), True),
        ((3, '<<'), (8, 8), True),
        ((3, '<<'), (8, 9), True),
        # n, *  no comparison
        ((2, '*'), (5, 6), True),
        ((2, '*'), (5, 3), True),
        ((2, '*'), (5, 5), True),
        ((2, '**'), (5, 6), True),
        ((2, '**'), (5, 3), True),
        ((2, '**'), (5, 5), True),
        # ><, 0
        (('>', 0), (6, 3), True),
        (('>', 0), (3, 3), False),
        (('>', 0), (3, 6), False),
        (('>>', 0), (6, 3), True),
        (('>>', 0), (3, 3), False),
        (('>>', 0), (3, 6), False),
        (('<', 0), (3, 6), True),
        (('<', 0), (3, 3), False),
        (('<', 0), (6, 3), False),
        (('<<', 0), (3, 6), True),
        (('<<', 0), (3, 3), False),
        (('<<', 0), (6, 3), False),
        # ><, n  no comparison
        (('>', 3), (8, 9), True),
        (('>', 3), (8, 8), True),
        (('>', 3), (8, 5), True),
        (('>>', 3), (8, 9), True),
        (('>>', 3), (8, 8), True),
        (('>>', 3), (8, 5), True),
        (('<', 3), (8, 6), True),
        (('<', 3), (8, 8), True),
        (('<', 3), (8, 9), True),
        (('<<', 3), (8, 6), True),
        (('<<', 3), (8, 8), True),
        (('<<', 3), (8, 9), True),
        # ><, ><
        (('>', '>'), (3, 3), True),
        (('>', '>'), (3, 4), False),
        (('>', '>'), (3, 1), False),
        (('>>>', '>>>'), (3, 3), True),
        (('>>>', '>>>'), (3, 4), False),
        (('>>>', '>>>'), (3, 1), False),
        (('<', '<'), (3, 3), True),
        (('<', '<'), (3, 4), False),
        (('<', '<'), (3, 1), False),
        (('<<<', '<<<'), (3, 3), True),
        (('<<<', '<<<'), (3, 4), False),
        (('<<<', '<<<'), (3, 1), False),
        (('>', '>>'), (3, 6), True),
        (('>', '>>'), (3, 3), False),
        (('>', '>>'), (3, 1), False),
        (('>>', '>>>'), (3, 6), True),
        (('>>', '>>>'), (3, 3), False),
        (('>>', '>>>'), (3, 1), False),
        (('<', '<<'), (5, 3), True),
        (('<', '<<'), (5, 5), False),
        (('<', '<<'), (5, 7), False),
        (('>>', '<<<'), (5, 3), True),
        (('>>', '<<<'), (5, 5), False),
        (('<<', '>>>'), (3, 5), True),
        (('<<', '>>>'), (3, 3), False),
        (('<<', '>>>'), (3, 1), False),
        # ><, *  no comparison
        (('>', '*'), (3, 3), True),
        (('>>', '***'), (3, 2), True),
        (('<<<', '**'), (2, 5), True),
        # *, 0
        (('***', 0), (4, 5), True),
        (('**', 0), (5, 4), True),
        (('****', 0), (5, 5), False),
        # *, n  no comparison
        (('***', 1), (4, 5), True),
        (('**', 2), (5, 4), True),
        (('****', -3), (5, 5), True),
        # *, ><  no comparison
        (('***', '<<'), (4, 5), True),
        (('**', '>'), (5, 4), True),
        (('****', '>>>'), (5, 5), True),
        # *, *
        (('*', '*'), (6, 6), True),
        (('*', '*'), (6, 5), False),
        (('**', '**'), (6, 6), True),
        (('**', '**'), (6, 5), False),
        (('**', '****'), (5, 7), True),
        (('**', '****'), (5, 5), False),
    )
)
def test_match_order(orders, resids, answer):
    order1, order2 = orders
    resid1, resid2 = resids
    match = do_links.match_order(order1, resid1, order2, resid2)
    assert match == answer


@pytest.mark.parametrize("order", (
    1.2, -3.9, None, True, False, {'a': 3},
    '><', '#>>', '#$H', '', [], np.nan,
))
def test_order_errors(order):
    with pytest.raises(ValueError):
        do_links._interpret_order(order)


@pytest.mark.parametrize("order, ref_order_type, ref_order_value", (
    (0, 'number', 0),
    (2, 'number', 2),
    (-3, 'number', -3),
    ('>', '><', 1),
    ('>>', '><', 2),
    ('>>>', '><', 3),
    ('<', '><', -1),
    ('<<', '><', -2),
    ('<<<', '><', -3),
    ('*', '*', 1),
    ('**', '*', 2),
    ('***', '*', 3),
))
def test_interpret_order(order, ref_order_type, ref_order_value):
    order_type, order_value = do_links._interpret_order(order)
    assert order_type == ref_order_type
    assert order_value == ref_order_value


def make_mol(mol_nodes, mol_edges=tuple(), **kwargs):
    mol = Molecule(**kwargs)
    mol.add_nodes_from(mol_nodes)
    mol.add_edges_from(mol_edges)
    return mol


def make_link(mol_nodes, mol_edges=tuple()):
    mol = Link()
    mol.add_nodes_from(mol_nodes)
    mol.add_edges_from(mol_edges)
    return mol


@pytest.mark.parametrize('mol_nodes, mol_edges, link_nodes, link_edges, expected_nodes, expected_edges', (
    (
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'}), (2, {'atomname': 'c'})],
        [(0, 1), (1, 2)],
        [[(0, {'atomname': 'a'}), (1, {'atomname': 'b'}), (2, {'atomname': 'c', 'replace': {'atomname': None}})]],
        [[(0, 1), (1, 2)]],
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'})],
        [(0, 1)],
    ),
    (
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'})],
        [(0, 1)],
        [[(0, {'atomname': 'a', 'replace': {'atomname': 'a1'}}), (1, {'atomname': 'b', 'replace': {'atomname': None}})]],
        [[(0, 1)]],
        [(0, {'atomname': 'a1'})],
        [],
    ),
    (
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'}), (2, {'atomname': 'a'})],
        [(0, 1), (1, 2)],
        [[(0, {'atomname': 'a', 'replace': {'atomname': 'a1'}}), (1, {'atomname': 'b', 'replace': {'atomname': None}})]],
        [[(0, 1)]],
        [(0, {'atomname': 'a1'}), (2, {'atomname': 'a1'})],
        [],
    ),
    (
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'}), (2, {'atomname': 'c'})],
        [(0, 1), (1, 2)],
        [
            [(0, {'atomname': 'a', 'replace': {'atomname': 'a1'}}), (1, {'atomname': 'b', 'replace': {'atomname': None}})],
            [(0, {'atomname': 'a1', 'replace': {'atomname': 'a2'}}), (1, {'atomname': 'b'})]
        ],
        [[(0, 1)], [(0, 1)]],
        [(0, {'atomname': 'a1'}), (2, {'atomname': 'c'})],
        [],
    ),
    (
        [(0, {'atomname': 'a'}), (1, {'atomname': 'b'}), (2, {'atomname': 'c'})],
        [(0, 1), (1, 2)],
        [
            [(0, {'atomname': 'a', 'replace': {'atomname': 'a1'}}), (1, {'atomname': 'b', 'replace': {'atomname': 'b1'}})],
            [(0, {'atomname': 'c', 'replace': {'atomname': 'c1'}}), (1, {'atomname': 'b1', 'replace': {'atomname': 'b2'}})]
        ],
        [[(0, 1)], [(0, 1)]],
        [(0, {'atomname': 'a1'}), (1, {'atomname': 'b2'}), (2, {'atomname': 'c1'})],
        [(0, 1), (1, 2)],
    ),
))
def test_link_processor(mol_nodes, mol_edges, link_nodes, link_edges,
                        expected_nodes, expected_edges):
    links = [make_link(nodes, edges) for nodes, edges in zip(link_nodes, link_edges)]
    ff = vermouth.forcefield.get_native_force_field('charmm')
    ff.links = links

    mol = make_mol(mol_nodes, mol_edges, force_field=ff)

    out = DoLinks().run_molecule(mol)
    assert dict(out.nodes(data=True)) == dict(expected_nodes)
    assert set(out.edges(data=False)) == set(expected_edges)


def modification(*names):
    mod = Modification(name=tuple(names))
    return mod


@pytest.mark.parametrize('mol_node, link_node, expected', (
    [{}, {}, True],  # No restrictions, match
    [{'modifications': [modification('A')]}, {}, True],  # No restrictions, match
    [{}, {'modifications': []}, True],  # No link modifications, no mol modifications, match
    [{}, {'modifications': None}, True],  # No link modifications, no mol modifications, match
    [{'modifications': [modification('A')]}, {'modifications': None}, False],  # No mods allowed, no match
    [{'modifications': [modification('A')]}, {'modifications': []}, False],  # No mods allowed, no match
    [{'modifications': []}, {'modifications': []}, True],  # No link modifications, no mol modifications, match
    [{'modifications': []}, {'modifications': None}, True],  # No link modifications, no mol modifications, match
    [{}, {'modifications': ['A']}, False],   # Link modifications, but no mol modifications, don't match
    [{'modifications': []}, {'modifications': 'A'}, False],  # Link modifications, but no mol modifications, don't match
    [{'modifications': [modification('A')]}, {'modifications': 'A'}, True],  # Link modifications and matching mol mods, match
    [{'modifications': [modification('A')]}, {'modifications': 'B'}, False],  # Link modifications but no matching mol mods, don't match
    [{'modifications': [modification('A', 'B')]}, {'modifications': 'B'}, False],  # Link modifications but no not all mol mods match, don't match
    [{'modifications': [modification('A'), modification('B')]}, {'modifications': 'B'}, False],  # Link modifications but no not all mol mods match, don't match
    [{'modifications': [modification('A')]}, {'modifications': Choice(['A', 'B'])}, True],  # Link modifications and matching mol mods, match
    [{'modifications': [modification('A', 'B')]}, {'modifications': ['A', 'B']}, True],  # Link modifications and matching mol mods, match
    [{'modifications': [modification('A', 'B')]}, {'modifications': ['B', 'A']}, True],  # Link modifications and matching mol mods, match
    [{'modifications': [modification('A'), modification('B')]}, {'modifications': ['A', 'B']}, True],  # Link modifications and matching mol mods, match
    [{'modifications': [modification('A'), modification('B')]}, {'modifications': 'A'}, False],  # Unmatched mod, no match
    [{'modifications': [modification('A'), modification('A')]}, {'modifications': 'A'}, True],  # Multiple mods, all matched
    [{'modifications': [modification('A'), modification('B')]}, {'modifications': Choice(['A', 'B'])}, True],  # Choice
    [{'modifications': [modification('A'), modification('A')]}, {'modifications': Choice(['A', 'B'])}, True],  # Choice
    [{'modifications': [modification('C'), modification('A')]}, {'modifications': Choice(['A', 'B'])}, False],  # Unmatched choice
))
def test_modification_matching(mol_node, link_node, expected):
    found = _atoms_match(mol_node, link_node)
    assert found == expected



# Copyright 2023 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Read RCSU Go model contact maps.
"""

def read_go_map(file_path):
    """
    Read a RCSU contact map from the c code as published in
    doi:10.5281/zenodo.3817447. The format requires all 
    contacts to have 18 columns and the first column to be 
    a capital R.

    Parameters
    ----------
    file_path: :class:`pathlib.Path`
        path to the contact map file

    Returns
    -------
    list(tuple)
        contact as chain id, res id, chain id, res id
    """
    with open(file_path, "r", encoding='UTF-8') as _file:
        contacts = []
        for line in _file:
            tokens = line.strip().split()
            if len(tokens) == 0:
                continue

            if tokens[0] == "R" and len(tokens) == 18:
                # this is a bad place to filter but follows
                # the old script
                if tokens[11] == "1" or (tokens[11] == "0" and tokens[14] == "1"):
                    # this is a OV or rCSU contact we take it
                    contacts.append((int(tokens[5]), tokens[4], int(tokens[9]), tokens[8]))

        if len(contacts) == 0:
            raise IOError("You contact map is empty. Are you sure it has the right formatting?")

    return contacts



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
"""

import networkx as nx
from ..molecule import Interaction
from ..processors.processor import Processor
from ..gmx.topology import Atomtype
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


class VirtualSiteCreator(Processor):
    """
    Create virtual-sites for the Martini Go model implementation or
    the specific water biasing options.

    See :mod:`vermouth.rcsu` for more details.

    Every molecule must have a moltype name under the "moltype" key of the
    molecule meta.

    See Also
    --------
    :class:`~vermouth.processors.name_moltype.NameMolType`
        Assign molecule type names to the molecules in a system.
    :func:`add_virtual_sites`
    """
    def __init__(self):
        self.system = None

    def run_molecule(self, molecule):
        moltype = molecule.meta.get('moltype')
        if not moltype:
            raise ValueError('The molecule does not have a moltype name.')

        if not self.system:
            raise ValueError('This processor requires a system.')
        molecule.citations.add('M3_GO')

        self.add_virtual_sites(molecule, prefix=moltype)

        return molecule

    def run_system(self, system):
        self.system = system
        LOGGER.info("Adding Virtual Sites to backbone beads.", type="step")
        super().run_system(system)

    def add_virtual_sites(self, molecule, prefix, backbone='BB', atomname='CA', charge=0):
        """
        Add the virtual sites for GoMartini in the molecule.

        One virtual site is added per backbone bead of the the Martini protein.
        Each virtual site copies the resid, resname, and chain of the backbone
        bead. It also copies the *reference* to the position array, so the virtual
        site position follows if the backbone bead is translated. The virtual sites
        are added *after* all the other atoms of the molecule, each in its own
        charge group, with "CA" as atomname, and a charge of 0. The atomname and
        charge can be set with the `atomname` and `charge` argument, respectively.

        The bead type of the virtual sites is names "<prefix>_<resid>". Where
        `prefix` is provided as an argument of the function, and is expected to be
        the molecule type name.

        Parameters
        ----------
        molecule: vermouth.molecule.Molecule
            The molecule to augment with virtual sites.
        prefix: str
            The prefix to use for bead type names. Usually the molecule type name.
        backbone: str
            The atomname of the backbone beads.
        atomname: str
            The atomname of the virtual sites.
        charge: float or int
            The charge of the virtual sites.
        """
        # If there are no atoms, then there is nothing to do. We can exit early,
        # avoiding to deal with empty iterators.
        if not molecule.nodes:
            return
        virtual_site_nodes = []
        virtual_sites = []
        atomtypes = {}
        new_node_id = max(molecule.nodes)
        charge_groups = nx.get_node_attributes(molecule, 'charge_group').values()
        new_charge_group = max(charge_groups) if charge_groups else 0
        for node_id, atom in molecule.nodes(data=True):
            if atom.get('atomname') == backbone:
                new_node_id += 1
                new_charge_group += 1

                virtual_site_nodes.append((new_node_id, {
                    'resid': atom['resid'],
                    '_old_resid': atom['_old_resid'],
                    'resname': atom['resname'],
                    'atype': '{}_{}'.format(prefix, atom['resid']),
                    'charge_group': new_charge_group,
                    'chain': atom['chain'],
                    'position': atom['position'],
                    'atomname': atomname,
                    'charge': charge,
                    'mass': 0.0,
                    'cgsecstruct': atom.get('cgsecstruct', None),
                }))
                virtual_sites.append(Interaction(
                    atoms=[new_node_id, node_id],
                    parameters=['1'],
                    meta={'go_vs': True, 'group': 'Virtual go site'},
                ))
                vs_params = Atomtype(node=new_node_id, molecule=molecule, sigma=0.0, epsilon=0.0, meta={})
                self.system.gmx_topology_params['atomtypes'].append(vs_params)

        molecule.add_nodes_from(virtual_site_nodes)


        molecule.interactions['virtual_sitesn'] += virtual_sites



# -*- coding: utf-8 -*-
# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the Go contact map reader.
"""
import pytest
import hypothesis
import hypothesis.strategies as st
from vermouth.rcsu.go_utils import _get_bead_size, _in_resid_region, get_go_type_from_attributes
from vermouth.tests.molecule_strategies import random_molecule

@pytest.mark.parametrize('atype, size', (
        ("SN4a", "small"),
        ("P3", "regular"),
        ("Tq4d", "tiny")
    ))
def test_get_bead_size(atype, size):
    assert size == _get_bead_size(atype)


@pytest.mark.parametrize('regions, resid, result', (
        # single region true
        ([(101, 120)], 115, True),
        # two regions true
        ([(101, 120), (180, 210)], 115, True),
        # two regions true; but inaccurate sorting
        ([(120, 101), (180, 210)], 115, True),
        # single region false
        ([(101, 120)], 15, False),
        # two regions false
        ([(101, 120), (180, 210)], 15, False),
    ))
def test_in_region(regions, resid, result):
    assert result == _in_resid_region(resid, regions)


@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.too_slow])
@hypothesis.given(random_molecule())
def test_get_go_type_from_attributes(mol):
    vs_node = len(mol.nodes)
    with pytest.raises(KeyError):
        next(get_go_type_from_attributes(mol, prefix="prefix", chain="A", resid=5))
    mol.add_node(vs_node, atype="prefix_0", chain="A", resid=5)
    found_atype = next(get_go_type_from_attributes(mol, prefix="prefix", chain="A", resid=5))
    assert found_atype == "prefix_0"



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the functions required to use DSSP.
"""

import os
import glob
import itertools

import networkx as nx
import numpy as np
import pytest

import vermouth
from vermouth.file_writer import DeferredFileWriter
from vermouth.forcefield import get_native_force_field
from vermouth.dssp import dssp
from vermouth.dssp.dssp import DSSPError, AnnotateDSSP
from vermouth.pdb.pdb import read_pdb
from vermouth.tests.datafiles import (
    PDB_PROTEIN,
    DSSP_OUTPUT,
    DSSP_SS_OUTPUT,
    PDB_ALA5_CG,
    PDB_ALA5,
)

DSSP_EXECUTABLE = os.environ.get("VERMOUTH_TEST_DSSP", "dssp")
SECSTRUCT_1BTA = list(
    "CEEEEETTTCCSHHHHHHHHHHHHTCCTTCCCSHHHHHHHHTTT"
    "SCSSEEEEEESTTHHHHTTTSSHHHHHHHHHHHHHTTCCEEEEEC"
)


# TODO: The code is very repetitive. There may be a way to refactor it with
# clever use of parametrize and fixtures.
class TestAnnotateResidues:
    """
    Tests for the :class:`dssp.AnnotateResidues` processor.
    """

    @staticmethod
    def build_molecule(nresidues):
        """
        Build a dummy molecule with the requested number of residues and 3
        atoms per residue.
        """
        molecule = vermouth.molecule.Molecule()
        residue_template = vermouth.molecule.Molecule()
        residue_template.add_nodes_from(
            (idx, {"chain": "", "atomname": str(idx), "resname": "DUMMY", "resid": 1})
            for idx in range(3)
        )
        for _ in range(nresidues):
            molecule.merge_molecule(residue_template)
        return molecule

    @staticmethod
    def sequence_from_mol(molecule, attribute, default=None):
        """
        Extract the content of an attribute for each node of a molecule.
        """
        return [node.get(attribute, default) for node in molecule.nodes.values()]

    def sequence_from_system(self, system, attribute, default=None):
        """
        Extract the content of an attribute for each node of a system.
        """
        return list(
            itertools.chain(
                *(
                    self.sequence_from_mol(molecule, attribute, default)
                    for molecule in system.molecules
                )
            )
        )

    @pytest.mark.parametrize("nres", (0, 1, 3, 10))
    def test_build_molecule(self, nres):
        """
        :meth:`build_molecule` and :meth:`sequence_from_mol` work as excpected.
        """
        expected_resid = list(itertools.chain(*([idx + 1] * 3 for idx in range(nres))))
        expected_atomname = ["0", "1", "2"] * nres
        expected_chain = [""] * (nres * 3)
        expected_resname = ["DUMMY"] * (nres * 3)
        molecule = self.build_molecule(nres)
        assert self.sequence_from_mol(molecule, "resid") == expected_resid
        assert self.sequence_from_mol(molecule, "resname") == expected_resname
        assert self.sequence_from_mol(molecule, "chain") == expected_chain
        assert self.sequence_from_mol(molecule, "atomname") == expected_atomname

    @pytest.fixture
    def single_mol_system(self):
        """
        Build a system with a single molecule that count 5 residues of 3 atoms
        each.
        """
        molecule = self.build_molecule(5)
        system = vermouth.system.System()
        system.molecules = [molecule]
        return system

    @pytest.fixture
    def multi_mol_system_irregular(self):
        """
        Build a system with 3 molecules having 4, 5, and 6 residues,
        respectively.
        """
        system = vermouth.system.System()
        system.molecules = [self.build_molecule(nres) for nres in (4, 5, 6)]
        return system

    @pytest.fixture
    def multi_mol_system_regular(self):
        """
        Build a system with 3 molecules having each 4 residues.
        """
        system = vermouth.system.System()
        system.molecules = [self.build_molecule(4) for _ in range(3)]
        return system

    @pytest.mark.parametrize(
        "sequence",
        (
            "ABCDE",
            ["A", "B", "C", "D", "E"],
            range(5),
        ),
    )
    def test_single_molecule(self, single_mol_system, sequence):
        """
        The simple case with a single molecule and a sequence of the right size
        works as expected.
        """
        expected = list(itertools.chain(*([element] * 3 for element in sequence)))
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_system(single_mol_system)
        found = self.sequence_from_system(single_mol_system, "test")
        assert found == expected

    @pytest.mark.parametrize(
        "sequence",
        (
            "ABCDEFGHIJKLMNO",
            list("ABCDEFGHIJKLMNO"),
            range(15),
        ),
    )
    def test_multi_molecules_diff_sizes(self, multi_mol_system_irregular, sequence):
        """
        The case of many protein of various sizes and a sequence of the right
        size works as expected.
        """
        expected = list(itertools.chain(*([element] * 3 for element in sequence)))
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_system(multi_mol_system_irregular)
        found = self.sequence_from_system(multi_mol_system_irregular, "test")
        assert found == expected

    @pytest.mark.parametrize(
        "sequence",
        (
            "ABCD",
            ["A", "B", "C", "D"],
            range(4),
        ),
    )
    def test_multi_molecules_cycle(self, multi_mol_system_regular, sequence):
        """
        The case with multiple molecules with all the same size and one
        sequence to repeat for each molecule works as expected.
        """
        expected = list(itertools.chain(*([element] * 3 for element in sequence)))
        expected = expected * 3
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_system(multi_mol_system_regular)
        found = self.sequence_from_system(multi_mol_system_regular, "test")
        assert found == expected

    def test_single_molecules_cycle_one(self, single_mol_system):
        """
        One molecule and a one element sequence to repeat over all residues of
        the molecule.
        """
        sequence = "A"
        expected = [sequence] * (5 * 3)
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_system(single_mol_system)
        found = self.sequence_from_system(single_mol_system, "test")
        assert found == expected

    def test_multi_molecules_cycle_one(self, multi_mol_system_irregular):
        """
        Many molecules and a one element sequence to repeat.
        """
        sequence = "A"
        expected = [sequence] * (15 * 3)
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_system(multi_mol_system_irregular)
        found = self.sequence_from_system(multi_mol_system_irregular, "test")
        assert found == expected

    @staticmethod
    @pytest.mark.parametrize(
        "sequence",
        (
            "ABC",  # Too short
            "ABCD",  # Too short, match the length of the first molecule
            "ABCDEFGHIFKLMNOPQRSTU",  # Too long
            "",  # Empty
        ),
    )
    def test_wrong_length(multi_mol_system_irregular, sequence):
        """
        Many molecule and a sequence that has the wrong length raises an error.
        """
        processor = dssp.AnnotateResidues("test", sequence)
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    @staticmethod
    @pytest.mark.parametrize(
        "sequence",
        (
            "ABC",  # Too short
            "ABCD",  # Too short, match the length of the first molecule
            "ABCDEFGHIFKLMNOPQRSTU",  # Too long
            "",  # Empty
            "ABCDEFGHIJKLMNO",  # Length of all the molecules, without filter
        ),
    )
    def test_wrong_length_with_filter(multi_mol_system_irregular, sequence):
        """
        Many molecules and a sequence that has the wrong length because of a
        molecule selector.
        """
        # We exclude the second molecule. The filter excludes it based on the
        # number of nodes, which is 15 because it has 5 residues with 3 nodes
        # each.
        processor = dssp.AnnotateResidues(
            "test",
            sequence,
            molecule_selector=lambda mol: len(mol.nodes) != (5 * 3),
        )
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    @staticmethod
    def test_empty_system_empty_sequence():
        """
        There are no molecules, but the sequence is empty.
        """
        system = vermouth.system.System()
        sequence = ""
        processor = dssp.AnnotateResidues("test", sequence)
        try:
            processor.run_system(system)
        except ValueError:
            pytest.fail("Should not have raised a ValueError.")

    @staticmethod
    def test_empty_system_error():
        """
        There are no molecules, but there is a sequence. Should raise an error.
        """
        system = vermouth.system.System()
        sequence = "not empty"
        processor = dssp.AnnotateResidues("test", sequence)
        with pytest.raises(ValueError):
            processor.run_system(system)

    @staticmethod
    def test_empty_with_filter(multi_mol_system_irregular):
        """
        There is a sequence, but no molecule are accepted by the molecule
        selector. Should raise an error.
        """
        sequence = "not empty"
        processor = dssp.AnnotateResidues(
            "test", sequence, molecule_selector=lambda mol: False
        )
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    def test_run_molecule(self, single_mol_system):
        """
        The `run_molecule` method works.
        """
        sequence = "ABCDE"
        expected = list(itertools.chain(*([element] * 3 for element in sequence)))
        processor = dssp.AnnotateResidues("test", sequence)
        processor.run_molecule(single_mol_system.molecules[0])
        found = self.sequence_from_system(single_mol_system, "test")
        assert found == expected

    def test_run_molecule_not_selected(self, single_mol_system):
        """
        The molecule selector works with `run_molecule`.
        """
        sequence = "ABCDE"
        processor = dssp.AnnotateResidues(
            "test", sequence, molecule_selector=lambda mol: False
        )
        processor.run_molecule(single_mol_system.molecules[0])
        found = self.sequence_from_system(single_mol_system, "test")
        assert vermouth.utils.are_all_equal(found)
        assert found[0] is None


@pytest.mark.parametrize(
    "input_file, expected",
    [
        (str(DSSP_OUTPUT), "".join(SECSTRUCT_1BTA)),
        (
            str(DSSP_SS_OUTPUT / "mini-protein1_betasheet.pdb.v2.2.1-3b2-deb_cv1.ssd"),
            "CEEEEEETTEEEEEECCCCCCTTCEEEEC",
        ),
        (
            str(DSSP_SS_OUTPUT / "mini-protein1_betasheet.pdb.v3.0.0-3b1-deb_cv1.ssd"),
            "CEEEEEETTEEEEEECCCCCCTTCEEEEC",
        ),
        (
            str(DSSP_SS_OUTPUT / "mini-protein2_helix.pdb.v2.2.1-3b2-deb_cv1.ssd"),
            "CCSHHHHHHHHHHCCCCHHHHHHHHHHHTSCHHHHHHHTCCC",
        ),
        (
            str(DSSP_SS_OUTPUT / "mini-protein2_helix.pdb.v3.0.0-3b1-deb_cv1.ssd"),
            "CCSHHHHHHHHHHCCCCHHHHHHHHHHHTSCHHHHHHHTCCC",
        ),
        (
            str(DSSP_SS_OUTPUT / "mini-protein3_trp-cage.pdb.v2.2.1-3b2-deb_cv1.ssd"),
            "CHHHHHHHTTGGGGTCCCCC",
        ),
        (
            str(DSSP_SS_OUTPUT / "mini-protein3_trp-cage.pdb.v3.0.0-3b1-deb_cv1.ssd"),
            "CHHHHHHHTTGGGGTCCCCC",
        ),
    ],
)
def test_read_dssp2(input_file, expected):
    """
    Test that :func:`vermouth.dssp.dssp.read_dssp2` returns the expected
    secondary structure sequence.
    """
    with open(input_file, encoding="utf-8") as infile:
        secondary_structure = dssp.read_dssp2(infile)
    assert "".join(secondary_structure) == expected


@pytest.mark.parametrize("savefile", [True, False])
def test_run_dssp(savefile, tmp_path):
    """
    Test that :func:`vermouth.molecule.dssp.dssp.run_dssp` runs as expected and
    generate a save file only if requested.
    """
    # The test runs twice, once with the savefile set to True so we test with
    # saving the DSSP output to file, and once with savefile set t False so we
    # do not generate the file. The "savefile" argument is set by
    # pytest.mark.parametrize.
    # The "tmp_path" argument is set by pytest and is the path to a temporary
    # directory that exists only for one iteration of the test.
    if savefile:
        path = tmp_path
    else:
        path = None
    system = vermouth.System()
    for molecule in read_pdb(str(PDB_PROTEIN)):
        system.add_molecule(molecule)
    secondary_structure = dssp.run_dssp(
        system, executable=DSSP_EXECUTABLE, savedir=path
    )

    # Make sure we produced the expected sequence of secondary structures
    assert secondary_structure == SECSTRUCT_1BTA

    # If we test with savefile, then we need to make sure the file is created
    # and its content corresponds to the reference (excluding the first lines
    # that are variable or contain non-essencial data read from the PDB file).
    # If we test without savefile, then we need to make sure the file is not
    # created.
    if savefile:
        DeferredFileWriter().write()
        assert path.exists()
        foundfile = list(path.glob('chain_*.ssd'))
        assert len(foundfile) == 1
        foundfile = foundfile[0]

        with open(foundfile, encoding="utf-8") as genfile, open(str(DSSP_OUTPUT), encoding="utf-8") as reffile:
            # DSSP 3 is outputs mostly the same thing as DSSP2, though there
            # are some differences in non significant whitespaces, and an extra
            # field header. We need to normalize these differences to be able
            # to compare.
            gen = "\n".join(
                [
                    line.strip().replace("            CHAIN", "")
                    for line in genfile.readlines()[6:]
                ]
            )
            ref = "\n".join([line.strip() for line in reffile.readlines()[6:]])
            assert gen == ref
    else:
        # Is the directory empty?
        assert not list(tmp_path.iterdir())


@pytest.mark.parametrize(
    "pdb, loglevel,expected",
    [
        (PDB_PROTEIN, 10, True),  # DEBUG
        (PDB_PROTEIN, 30, False),  # WARNING
        # Using a CG pdb will cause a DSSP error, which should preserve the input
        (PDB_ALA5_CG, 10, True),  # DEBUG
        (PDB_ALA5_CG, 30, True),  # WARNING
    ],
)
def test_run_dssp_input_file(tmp_path, caplog, pdb, loglevel, expected):
    """
    Test that the DSSP input file is preserved (only) in the right conditions
    """
    caplog.set_level(loglevel)
    system = vermouth.System()
    for molecule in read_pdb(str(pdb)):
        system.add_molecule(molecule)
    os.chdir(tmp_path)
    try:
        dssp.run_dssp(system, executable=DSSP_EXECUTABLE)
    except DSSPError:
        pass
    if expected:
        target = 1
    else:
        target = 0
    matches = glob.glob("dssp_in*.pdb")
    assert len(matches) == target, matches
    if matches:
        # Make sure it's a valid PDB file. Mostly anyway.
        list(read_pdb(matches[0]))


@pytest.mark.parametrize('ss_struct, expected', (
    (list('ABCDE'), list('ABCDE')),
    (list('AB DE'), list('ABCDE')),
    ([['A'], ['B'], ['C'], ['F'], ['G']], list('ABCFG')),
    ([['A'], [' '], ['E'], ['F'], [' ']], list('ACEFC')),
))
def test_mdtraj(monkeypatch, ss_struct, expected):
    # We don't want to test mdtraj.compute_dssp, so mock it.
    compute_dssp = lambda *_, **__: np.array(ss_struct)
    monkeypatch.setattr(vermouth.dssp.dssp.mdtraj, "compute_dssp", compute_dssp)
    system = vermouth.System()
    for molecule in read_pdb(str(PDB_ALA5)):
        system.add_molecule(molecule)

    processor = AnnotateDSSP(executable=None)
    processor.run_system(system)

    found = []
    for mol in system.molecules:
        residues = mol.iter_residues()
        for residue in residues:
            found.append(mol.nodes[residue[0]]['secstruct'])

    assert found == expected


def test_cterm_atomnames():
    nodes = [
        dict(
            resname="ALA",
            atomname="N",
            element="N",
            resid=1,
            chain="",
            position=np.array([9.534, 5.359, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="CA",
            element="C",
            resid=1,
            chain="",
            position=np.array([10.190, 6.661, -0.000]),
        ),
        dict(
            resname="ALA",
            atomname="C",
            element="C",
            resid=1,
            chain="",
            position=np.array([11.706, 6.515, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="O",
            element="O",
            resid=1,
            chain="",
            position=np.array([12.232, 5.403, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="CB",
            element="C",
            resid=1,
            chain="",
            position=np.array([9.733, 7.484, 1.196]),
        ),
        dict(
            resname="ALA",
            atomname="H",
            element="H",
            resid=1,
            chain="",
            position=np.array([10.101, 4.523, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="HA",
            element="H",
            resid=1,
            chain="",
            position=np.array([9.914, 7.191, -0.912]),
        ),
        dict(
            resname="ALA",
            atomname="1HB",
            element="H",
            resid=1,
            chain="",
            position=np.array([10.231, 8.454, 1.181]),
        ),
        dict(
            resname="ALA",
            atomname="2HB",
            element="H",
            resid=1,
            chain="",
            position=np.array([8.654, 7.630, 1.147]),
        ),
        dict(
            resname="ALA",
            atomname="3HB",
            element="H",
            resid=1,
            chain="",
            position=np.array([9.987, 6.960, 2.116]),
        ),
        dict(
            resname="ALA",
            atomname="N",
            element="N",
            resid=2,
            chain="",
            position=np.array([12.404, 7.646, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="CA",
            element="C",
            resid=2,
            chain="",
            position=np.array([13.862, 7.646, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="C",
            element="C",
            resid=2,
            chain="",
            position=np.array([14.413, 9.066, 0.000]),
        ),
        dict(
            resname="ALA",
            atomname="O1",
            element="O",
            resid=2,
            chain="",
            position=np.array([14.462, 9.691, 1.023]),
        ),
        dict(
            resname="ALA",
            atomname="O2",
            element="O",
            resid=2,
            chain="",
            position=np.array([14.798, 9.560, -1.023]),
        ),
        dict(
            resname="ALA",
            atomname="CB",
            element="C",
            resid=2,
            chain="",
            position=np.array([14.392, 6.868, -1.196]),
        ),
        dict(
            resname="ALA",
            atomname="H",
            element="H",
            resid=2,
            chain="",
            position=np.array([11.912, 8.528, -0.000]),
        ),
        dict(
            resname="ALA",
            atomname="HA",
            element="H",
            resid=2,
            chain="",
            position=np.array([14.212, 7.162, 0.912]),
        ),
        dict(
            resname="ALA",
            atomname="1HB",
            element="H",
            resid=2,
            chain="",
            position=np.array([15.482, 6.878, -1.181]),
        ),
        dict(
            resname="ALA",
            atomname="2HB",
            element="H",
            resid=2,
            chain="",
            position=np.array([14.038, 5.839, -1.147]),
        ),
        dict(
            resname="ALA",
            atomname="3HB",
            element="H",
            resid=2,
            chain="",
            position=np.array([14.038, 7.331, -2.116]),
        ),
    ]
    edges = [
        (0, 1),
        (0, 5),
        (1, 2),
        (1, 4),
        (1, 6),
        (2, 3),
        (4, 7),
        (4, 8),
        (4, 9),
        (2, 10),
        (10, 11),
        (10, 16),
        (11, 12),
        (11, 15),
        (11, 17),
        (12, 13),
        (12, 14),
        (15, 18),
        (15, 19),
        (15, 20),
    ]
    ff = get_native_force_field("charmm")
    mol = vermouth.molecule.Molecule(force_field=ff)
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from(edges)
    system = vermouth.system.System(force_field=ff)
    system.add_molecule(mol)
    vermouth.processors.RepairGraph().run_system(system)
    vermouth.processors.CanonicalizeModifications().run_system(system)
    dssp_out = dssp.run_dssp(system, executable=DSSP_EXECUTABLE)
    assert dssp_out == list("CC")


@pytest.mark.parametrize('sequence, expected', [
    ('H', '3'),
    ('HH', '33'),
    ('CHH', 'C33'),
    ('HHHHHH', '113322'),
    ('EHHHHHHC', 'E113322C'),
    ('HHHHHHHHH', '1111H2222'),
    ('CHHHHHHHHHC', 'C1111H2222C'),
    ('CHHHHEHHHHC', 'C3333E3333C'),
])
def test_convert_dssp_to_martini(sequence, expected):
    found = dssp.convert_dssp_to_martini(sequence)
    assert expected == found
